import zope.sqlalchemy
from pyramid.config import Configurator
from pyramid.events import NewRequest
from pyramid.renderers import JSONP
from pyramid.request import Request
from pyramid.view import view_config, view_defaults
from pyramid_beaker import session_factory_from_settings

from workspaces.capability.services.capability_info import CapabilityInfo
from workspaces.capability.services.capability_service import CapabilityService
from workspaces.system.schema import get_engine, get_session_factory
from workspaces.workflow.services.workflow_service import WorkflowServiceRESTClient


# Copied from here: https://stackoverflow.com/questions/21107057/pyramid-cors-for-ajax-requests
def add_cors_headers_response_callback(event):
    def cors_headers(request, response):
        response.headers.update(
            {
                "Access-Control-Allow-Origin": "*",
                "Access-Control-Allow-Methods": "POST,GET,DELETE,PUT,OPTIONS",
                "Access-Control-Allow-Headers": "Origin, Content-Type, Accept, Authorization",
                "Access-Control-Allow-Credentials": "true",
                "Access-Control-Max-Age": "1728000",
                "Content-Type": "application/json",
            }
        )

    event.request.add_response_callback(cors_headers)


# ---------------------------------------------------------
#
#    S E R V I C E S
#
# ---------------------------------------------------------


def lookup_request(request):
    return request.capability_info.lookup_capability_request(
        request.matchdict["request_id"]
    )


@view_defaults(route_name="capability_request", renderer="json")
class CapabilityRestService:
    def __init__(self, request: Request):
        self.request = request

    @view_config(request_method="POST", renderer="json")
    def create(self):
        # create a capability request for this ... request
        req = self.request.json_body
        request = self.request.capabilities.create_request(req["capability"])
        return request

    @view_config(
        request_method="POST", route_name="submit_capability_request", renderer="json"
    )
    def submit(self):
        # 1. Submit the request to the service
        execution = self.request.capabilities.run_capability(self.request.context)

        # 2. Return something we can listen for
        return execution


# ---------------------------------------------------------
#
#    F R A M E W O R K    S E T U P
#
# ---------------------------------------------------------


def get_tm_session(session_factory, transaction_manager):
    """
    Enable Zope's transaction manager on our session
    :param session_factory:
    :param transaction_manager:
    :return:
    """
    dbsession = session_factory()
    zope.sqlalchemy.register(dbsession, transaction_manager=transaction_manager)
    return dbsession


# ---------------------------------------------------------
#
#    M A I N    E N T R Y P O I N T
#
# ---------------------------------------------------------


def main(global_config, **settings):
    with Configurator(settings=settings) as config:
        config.add_subscriber(add_cors_headers_response_callback, NewRequest)
        session_factory = session_factory_from_settings(settings)
        config.set_session_factory(session_factory)
        config.add_renderer("jsonp", JSONP(param_name="callback"))

        settings["tm.manager_hook"] = "pyramid_tm.explicit_manager"

        # use pyramid_tm to hook the transaction lifecycle to the request
        config.include("pyramid_tm")

        # use pyramid_retry to retry a request when transient exceptions occur
        config.include("pyramid_retry")

        session_factory = get_session_factory(get_engine())
        config.registry["dbsession_factory"] = session_factory

        # make capability_info available for use in Pyramid
        config.add_request_method(
            # r.tm is the transaction manager used by pyramid_tm
            lambda request: CapabilityInfo(get_tm_session(session_factory, request.tm)),
            "capability_info",
            reify=True,
        )
        # make workflow_info available for use in Pyramid
        config.add_request_method(
            # r.tm is the transaction manager used by pyramid_tm
            lambda request: WorkflowServiceRESTClient(),
            "workflow_service",
            reify=True,
        )
        # make capability_service available for use in Pyramid
        config.add_request_method(
            lambda r: CapabilityService(r.capability_info, r.workflow_service),
            "capabilities",
            reify=True,
        )

        # add some routes
        config.add_route("capability_request", "/capability/request")
        config.add_route(
            "submit_capability_request",
            "/capability/request/{request_id}/submit",
            factory=lookup_request,
        )

        config.include("pyramid_beaker")
        config.scan(".")
        return config.make_wsgi_app()
